//Ethan Alexander Shulman 2022 "Metal Spheres" May 4, 2022

//scene rendering options
#define VIEW_STEPS 60
#define VIEW_STEP_DITHER .1
#define VIEW_EPSILON 1e-3
#define VIEW_EPSILON_MUL 1.12
#define LIGHTING_STEPS 60
#define LIGHTING_STEP_DITHER .1
#define LIGHTING_EPSILON 1e-2
//#define COLOR_MAX 100.
#define DENOISE_RANGE 1
#define DENOISE_NORMAL_THRESHOLD .98
#define DENOISE_DEPTH_THRESHOLD 1e2
#define LOW_DISCREPANCY 1
//#define ANIMATION_TIME 30.
#define ANIMATION_SAMPLES 100
#ifdef ANIMATION_TIME
out float time;
#else
float time = 25.;
#endif

//constants
#define PI 3.14159265



float twave(float x) {
	return 1.-abs(fract(x)-.5)*2.;
}

mat2 r2d(float a) {
	float c = cos(a), s = sin(a);
	return mat2(c,s,-s,c);
}

uvec4 hash(uvec4 seed) {
    uvec4 h = (0x6A7F8FAAu^seed)*0x01000193u;
    h = ((h.wxyz>>3u)^h^seed.yzwx)*0x01000193u;
    h = ((h.zwxy>>8u)^h^seed.wxyz)*0x01000193u;
    return h^(h>>11u);
}
#define I2F (1./float(0xFFFFFFFFu))

float vvoronoi(vec3 p, vec3 s, float m) {
	return min(length(fract(1000.+p+cos(s)*m)-.5),length(fract(1000.+p-.5+sin(s*.8634)*m)-.5))*1.15;
}


/*returns camera data as mat3 in the following structure {
vec3 position;
float roll, pitch, yaw;
float fieldOfViewTangent, viewNear, orbitDistance;*/
mat3 camera() {
	float et = time;
	return mat3(6.*4.+2.+floor(et/5.)*4.+max(0.,mod(et,5.)-4.)*4.,1,-5,
			0,-.05,0,
			3.,2.5,0);
}


//distance dithering for volumetrics
#define START_VOLUME(posSeed) vec2 opacityRand = fract(baseRand.zw+128.*dot(abs(posSeed),baseRand.zxy)*baseRand.xy)
#define VOLUME(dst,skip,alpha,anisotropy) max(dst,-1e3+(skip+1e3)*step(alpha+nsampling*anisotropy,opacityRand.x))

//dither materials for chromatic separation
#define START_CHROMA(posSeed) float chromaFrame = mod(baseRand.x+156.*dot(abs(posSeed),baseRand.zwy),3.)
#define CHROMA(rc,gc,bc) if (chromaFrame < 1.) {rc;}\
else if (chromaFrame < 2.) {gc;}\
else {bc;}



//functions defining scene
//returns vec2(distance, materialId) of nearest surface
vec2 geometry(vec3 p) {
	//spheres
	vec3 rp = p;
	rp.y -= 1.;
	rp.x = mod(p.x,4.)-2.;
	vec2 geo = vec2(length(rp)-1.,floor(mod(p.x/4.,6.)));

	//floor plane
	float fdst = p.y;
	if (fdst < geo.x) geo = vec2(fdst,-1.);
	
	return geo;
}

/*returns material as mat3 in the following structure {
vec3 color;
float roughness, emission, opacity;
float refractionIndex;*/
mat3 material(vec3 p, vec3 n, vec3 rd, float mat) {
	float fres = -dot(rd,n);
	if (mat == 0.) {
		//stainless steel
		return mat3(vec3(.7), pow(baseRand.w,2.)*.1,0,1, 1,0,0);
	}
	if (mat == 1.) {
		//copper
		return mat3(vec3(255,136,102)*.7/255., .045,0,1, 1,0,0);
	}
	if (mat == 2.) {
		//gold
		return mat3(pow(vec3(209,181,98)/255.,vec3(1.7)), .03,0,1, 1,0,0);
	}
	if (mat == 3.) {
		//silver
		float v = vvoronoi(p*6.,p.zxy*6.*vec3(10,11,12),.5);
		return mat3(vec3(.5), .03+.1*v,0,1, 1,0,0);
	}
	if (mat == 4.) {
		//black zirconium
		return mat3(vec3(.1), fract(p.y*100.+1000.)*fract(p.x*146.8273+p.z*123.786)*.1,0,1, 1,0,0);
	}
	if (mat == 5.) {
		//anodized titanium
		mat3 tm;
		START_CHROMA(p);
		CHROMA(tm = mat3(1,0,0, .01,0,1, 1,0,0),
		 tm = mat3(0,1,0, .04,0,1, 1,0,0),
		 tm = mat3(0,0,1, .1,0,1, 1,0,0));
		return tm;
	}

	return mat3(vec3(.9), 1.,0,1, 1,0,0);
}

//returns background color at point and direction
vec3 background(vec3 p, vec3 d) {
	return vec3(1);
}
vec3 backgroundLighting(vec3 p, vec3 d) {
	return vec3(.01)+vec3(4)*max(0.,1.-length(d-normalize(vec3(1))));
}

